/* daemonize.c

    Taken from the book "C en action" (O'Reilly Editions)

    Adapted by Clément NOTIN for Erasme Centre Multimédia
    Summer 2007
*/

#include <stdio.h> //perror()
#include <stdlib.h> //exit()
#include <unistd.h> //fork(), setsid()...
#include <fcntl.h>   //FD settings
#include <string.h> //strlen()
#include <sys/stat.h> //umask()
#include <pwd.h> // passwd parsing
#include <grp.h> // group parsing

extern char *PID_FILE;
extern char *USER;
extern char *GROUP;

//##########################################################
//function from the book "C en action" (O'Reilly Editions)
//used to properly daemonize the program
void daemonize (void)
{
    int pid;
    int i;
    int fd;
    char str[12];
    //Escape from the father process
    pid = fork ();
    if (pid < 0)
    {
        perror ("Fork 1 error ");
        exit (1);
    }
    if (pid > 0)
    exit (0);

    setsid ();
    //Escape from the control terminal inherited from new session created by setsid
    pid = fork ();
    if (pid < 0)
    {
        perror ("Fork 2 error ");
        exit (1);
    }
    if (pid > 0)
    exit (0);
    //closing all file descriptors
    for (i = getdtablesize (); i >= 0; --i)
    close (i);
  
    //will be used to know if another instance is running, if this is the case it will exit gracely
    fd = open (PID_FILE,
                O_RDWR | O_CREAT, 0640);

    if (fd < 0)                   //file creating problem
    { perror("Pidfile problem "); exit (1); }
   
    //change program's group and user
    struct group *group;
    if(!(group=getgrnam(GROUP)))
    { fprintf(stderr, "Error : it looks like %s is not a valid group, be careful about the case !\n", GROUP); }
    setgid(group->gr_gid);
    
    struct passwd *user;
    if(!(user=getpwnam(USER)))
    { fprintf(stderr, "Error : it looks like %s is not a valid user, be careful about the case !\n", USER); }
    setuid(user->pw_uid);
    

    //Reopen file descriptor 0,1,2
    i = open ("/dev/null", O_RDWR);
    dup (i);
    dup (i);

    umask (022);
    chdir ("");

    if (lockf (fd, F_TLOCK, 0) < 0)       //test and locking of the file
    exit (0);
    snprintf (str, 12, "%d\n", getpid());
    write (fd, str, strlen (str));

}
